#ifndef _BARECTF_PLATFORM_PARALLELLA_H
#define _BARECTF_PLATFORM_PARALLELLA_H

/*
 * barectf Parallella platform
 *
 * Copyright (c) 2015 EfficiOS Inc. and Linux Foundation
 * Copyright (c) 2015 Philippe Proulx <pproulx@efficios.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "barectf.h"

/**
 * Initializes the platform.
 */
int tracing_init(void);

/**
 * Returns the barectf context to be used with tracing functions.
 */
struct barectf_default_ctx *tracing_get_barectf_ctx(void);

/**
 * Resets the tracing clock to an absolute 0.
 *
 * This should be used immediately after a synchronization operation
 * on all executed Epiphany cores.
 */
void tracing_reset_clock(void);

/**
 * Finalizes the platform.
 */
void tracing_fini(void);

#endif /* _BARECTF_PLATFORM_PARALLELLA_H */
